/******************************************************************************/
/* cmfutils.c      CICS on Open Systems Monitor Data Formatter   Rags Gardner */
/*                                                                            */
/* Rags Int., Inc.                                              December 1999 */
/* (c) RISC Application Growth Solutions International, Incorporated  1999    */
/*             (817) 267-2554     rags@compuserve.com                         */
/*                                                                            */
/* MonExpand()   formats field values in a CMF record for printing            */
/* MonCompare()  compares a field in a CMF record to user Include/Exclude     */
/* MonTitle()    formats title text based on user selected fields             */
/* insertRow()   inserts a new row for summary accumulation routines          */
/******************************************************************************/
#ifdef WIN32
   #include <windows.h>                                            /* windows */
#endif
#include <stdio.h>                                       /* Unix standard I/O */
#include <time.h>                                      /* Unix time functions */
#include <sys/types.h>                                 /* standard data types */
#include <memory.h>                                                 /* memory */
#include <string.h>                                                /* strings */
#include <cics_types.h>                            /* CICS utility data types */
#include <cicstype.h>                              /* CICS portable datatypes */
#include <cicsmfmt.h>                             /* monitor formatter header */
#include <cics_emp_tdq.h>                    /* monitor data field data types */
#include "cmfreport.h"                                    /* cmfreport header */

/*---------------------------------*/
/* Global Storage                  */
/*---------------------------------*/
struct CICS_EMP_Record                        /* User CMF clocks and counters */
{
   cics_ulong_t   Clk_Counter1;                            /* Clock 1 counter */
   cics_ulong_t   Clk_Time1;                               /* Clock 1 value   */
   cics_ulong_t   Clk_Counter2;                            /* Clock 2 counter */
   cics_ulong_t   Clk_Time2;                               /* Clock 2 value   */
   cics_slong_t   Counter1r;                               /* Counter 1       */
   cics_slong_t   Counter2r;                               /* Counter 2       */
   cics_slong_t   Counter3r;                               /* Counter 3       */
};
extern int    GMT;                        /* show time stamps in local or GMT */
extern int    PrtCount;              /* print counters associated with clocks */
extern int    SummaryRF;                    /* accumulate summary report flag */
extern int    TimeInSec;      /* timers reported in seconds.milliseconds flag */
extern int    MonFields;             /* number of monitor field types defined */
extern int    FieldCount;              /* number of fields in the select list */
extern int    LineNo;                                   /* current line count */
extern char   ProgName[];                                /* this program name */
extern struct FieldDesc_t FormatMap[];     /* cmf field formatting structures */
extern struct SummaryTable_t Table;                 /* the accumulation table */

/******************************************************************************/
/* MonExpand()                         Expand print line for a detail record  */
/******************************************************************************/
int MonExpand  (CICS_EMP_TDQ_RecordDetails_t *RecordDetails,
                struct FieldDesc_t           *Fieldp,
                char                          fieldValue[],
                struct SummaryRow_t          *row)
{
  struct   CICS_EMP_Record    *empPtr;    /* CMF user detail record structure */
  long     ElapsedTime;                        /* time calculation work areas */
  int      total_hours, total_minutes, total_seconds;
  int      hours, minutes, seconds, milliseconds;
  int      i, fldSize;                                       /* work integers */
  char     wrkbuf[64];                           /* string creation work area */
  char     HH[4];                                                    /* hours */
  char     MM[4];                                                  /* minutes */
  char     SS[4];                                                  /* seconds */
  char     MS[5];                                             /* milliseconds */
  float    fpointnum;                                 /* big number work area */
  struct   tm TimeData;                                         /* time value */
  time_t   Time;                                            /* time work area */

    /*-------------------------*/
    /* Process Invalid Fields  */
    /*-------------------------*/
 /* if (RecordDetails == NULL)              */    /* Invalid Field Selected ? */
 /* {                                       */
 /*     sprintf(fieldValue, "<Invalid>");   */
 /*     return(1);                          */
 /* }                                       */

    /*-------------------------*/
    /* Process Output Width    */
    /*-------------------------*/
    if (Fieldp->outSize < 1 || Fieldp->outSize > PRT_MAX+1 )
    {
        sprintf(fieldValue, "<lth-err>");
        return(1);
    } else {
        fldSize=Fieldp->outSize;          /* get the output report field size */
    }

    /*----------------------------------*/
    /* Process CMF Special Data Fields  */
    /*----------------------------------*/
    if (RecordDetails->FieldID == CICS_EMP_FID_TASK_FLAGS)
    {                                  /* Special case - bit flags  u_long_t  */
        sprintf(wrkbuf, "%8.8X", RecordDetails->Value.Counter);
#ifdef DEBUG2
        fprintf(stderr, "[%s]: Type(%i) FID_TASK_FLAGS",
            ProgName, RecordDetails->Type);
#endif
        sprintf(fieldValue, "%*.*s", fldSize, fldSize, wrkbuf);  /* format it */

            /* save the displayable value if field is a major or minor column */
        if (SummaryRF == TRUE) {
           if (Table.Major->empNum == RecordDetails->FieldID)
              strcpy(row->MajorValue, fieldValue);
           if (Table.Minor != NULL)
              if (Table.Minor->empNum == RecordDetails->FieldID)
                 strcpy(row->MinorValue, fieldValue);
        }
        return(0);
    }

    /*----------------------------------------*/
    /* Process CMF User 2 Clocks & 3 Counters */
    /*----------------------------------------*/
    if (RecordDetails->FieldID == CICS_EMP_FID_USER_SPACE
        && RecordDetails->Type == CICS_EMP_TDQ_DETAIL_TYPE_USER
        && Fieldp->usrType == none)
    {
     char   wrkbuf2[64];                  /* second string creation work area */
     char   wrkbuf3[64];                  /* second string creation work area */
       empPtr = (struct CICS_EMP_Record *)&RecordDetails->Value;
       sprintf(wrkbuf, "EMP(%i) No User Data(%i)",
                       RecordDetails->FieldID, RecordDetails->FieldSize);
       if (RecordDetails->FieldSize != 0) {
            memset(wrkbuf2, '\0', sizeof(wrkbuf2));
            memset(wrkbuf3, '\0', sizeof(wrkbuf3));
            ElapsedTime=empPtr->Clk_Time1/1000000;
            milliseconds=empPtr->Clk_Time1%1000000;   
            milliseconds=milliseconds/1000;   
            if (ElapsedTime || milliseconds)
                sprintf(wrkbuf2, "%8d.%.03d", ElapsedTime, milliseconds);
            if (wrkbuf2[0]=='0') wrkbuf2[0]=' ';
            sprintf(wrkbuf3, "%12.12s", wrkbuf2);
            strcpy(wrkbuf, wrkbuf3);

            sprintf(wrkbuf2, "%d", empPtr->Clk_Counter1);
            if (wrkbuf2[0]=='0') wrkbuf2[0]=' ';
            sprintf(wrkbuf3, " %6.6s", wrkbuf2);
            strcat(wrkbuf, wrkbuf3);

            memset(wrkbuf2, '\0', sizeof(wrkbuf2));
            ElapsedTime=empPtr->Clk_Time2/1000000;
            milliseconds=empPtr->Clk_Time2%1000000;   
            milliseconds=milliseconds/1000;   
            if (ElapsedTime || milliseconds)
                sprintf(wrkbuf2, "%8d.%.03d", ElapsedTime, milliseconds);
            if (wrkbuf2[0]=='0') wrkbuf2[0]=' ';
            sprintf(wrkbuf3, " %12.12s", wrkbuf2);
            strcat(wrkbuf, wrkbuf3);

            sprintf(wrkbuf2, "%d", empPtr->Clk_Counter2);
            if (wrkbuf2[0]=='0') wrkbuf2[0]=' ';
            sprintf(wrkbuf3, " %6.6s", wrkbuf2);
            strcat(wrkbuf, wrkbuf3);

            sprintf(wrkbuf2, "%d", empPtr->Counter1r);
            if (wrkbuf2[0]=='0') wrkbuf2[0]=' ';
            sprintf(wrkbuf3, " %6.6s", wrkbuf2);
            strcat(wrkbuf, wrkbuf3);

            sprintf(wrkbuf2, "%d", empPtr->Counter2r);
            if (wrkbuf2[0]=='0') wrkbuf2[0]=' ';
            sprintf(wrkbuf3, " %6.6s", wrkbuf2);
            strcat(wrkbuf, wrkbuf3);

            sprintf(wrkbuf2, "%d", empPtr->Counter3r);
            if (wrkbuf2[0]=='0') wrkbuf2[0]=' ';
            sprintf(wrkbuf3, " %6.6s", wrkbuf2);
            strcat(wrkbuf, wrkbuf3);
       }
       sprintf(fieldValue, "%*.*s", fldSize, fldSize, wrkbuf);
       return(0);
    }

    /*-----------------------------------*/
    /* Process CMF Standard Data Fields  */
    /*-----------------------------------*/
    switch(RecordDetails->Type)          /* Format a field based on it's type */
    {
        case CICS_EMP_TDQ_DETAIL_TYPE_STRING:
            /*-----------------------------------*/
            /* Character Strings ...             */
            /*-----------------------------------*/
#ifdef DEBUG2
            fprintf(stderr, "[MonExpand]: STRING(%i) %s\n",
                    fldSize, Fieldp->fieldName);
#endif
            memset(wrkbuf, ' ', sizeof(wrkbuf));             /* set to blanks */
            memcpy(wrkbuf, RecordDetails->Value.String,
                   RecordDetails->FieldSize);                /* get the array */
            wrkbuf[RecordDetails->FieldSize]='\0';  /* add string termination */
            sprintf(fieldValue, "%*.*s", fldSize, fldSize, wrkbuf); /* return */

            if (SummaryRF == TRUE) {               /* is this a summary run ? */
               /*-----------------------------------------------------*/
               /* save the string if field is a major or minor column */
               /*-----------------------------------------------------*/
               if (Table.Major->empNum == RecordDetails->FieldID)
                  strcpy(row->MajorValue, fieldValue);
               if (Table.Minor != NULL)
                  if (Table.Minor->empNum == RecordDetails->FieldID)
                     strcpy(row->MinorValue, fieldValue);
            }
            break;

        case CICS_EMP_TDQ_DETAIL_TYPE_COUNTER:              /* CICS_EMP_Count */
        case CICS_EMP_TDQ_DETAIL_TYPE_OS_COUNT:          /* CICS_EMP_OS_Count */
            /*-----------------------------------*/
            /* CICS & System Event Counters      */
            /*-----------------------------------*/
#ifdef DEBUG2
            fprintf(stderr, "[MonExpand]: COUNTER(%i) %s\n",
                    fldSize, Fieldp->fieldName);
#endif
            if (RecordDetails->FieldID != CICS_EMP_FID_TCASEQ_NO
                  || RecordDetails->FieldID != CICS_EMP_FID_TASK_FLAGS) {
               Fieldp->ValueCtr = Fieldp->ValueCtr
                     + RecordDetails->Value.Counter; /* update counter totals */

               if (SummaryRF == TRUE) {           /* is this a summary run ? */
                  for(i=0; i<FieldCount; i++)  /* capture this row's counter */
                     if (RecordDetails->FieldID
                          == row->ItemS[i].Fieldp->empNum) {
                        row->ItemS[i].RowCount = RecordDetails->Value.Counter;
                        break;
                     }
               }
            }
            sprintf(wrkbuf, "%ld", RecordDetails->Value.Counter);
            sprintf(fieldValue, "%*.*s", fldSize, fldSize, wrkbuf); /* return */
            break;

        case CICS_EMP_TDQ_DETAIL_TYPE_TIME_STAMP:
            /*----------------------------------------------------*/
            /* Basic Time Stamps ...                              */
            /* Timer in TOD format: CICS_EMP_Time_t               */
            /* convert to TOD hh:mm:ss.ms using ..time() function */ 
            /* The timer accuracy for this field is nanoseconds   */
            /*----------------------------------------------------*/
#ifdef DEBUG2
            fprintf(stderr, "[MonExpand]: TIME_STAMP(%i) %s\n",
                    fldSize, Fieldp->fieldName);
#endif
            Time = RecordDetails->Value.TimeStamp.tv_sec;
            Time += RecordDetails->Value.TimeStamp.tv_usec/1000000;
            if (GMT) {
               memcpy(&TimeData, gmtime(&Time), sizeof(struct tm));
            } else {
               memcpy(&TimeData, localtime(&Time), sizeof(struct tm));
            }
            sprintf(wrkbuf, "%02d:%02d:%02d.%.03d",
               TimeData.tm_hour, TimeData.tm_min, TimeData.tm_sec,
               RecordDetails->Value.TimeStamp.tv_usec/1000);
            sprintf(fieldValue, "%*.*s", fldSize, fldSize, wrkbuf); /* return */
            break;

        case CICS_EMP_TDQ_DETAIL_TYPE_OS_CLOCK:
            /*---------------------------------------------------*/
            /* Doubleword Time Stamps ...                        */
            /* Time value in two words format: CICS_EMP_OS_Clock */
            /* convert this to hh:mm:ss.ms or ssssssss.ms        */ 
            /* The timer accuracy for this field is microseconds */
            /*---------------------------------------------------*/
#ifdef DEBUG2
            fprintf(stderr, "[MonExpand]: OS_CLOCK(%i) %s\n",
                    fldSize, Fieldp->fieldName);
#endif
            ElapsedTime=RecordDetails->Value.TimeStamp.tv_sec;
            milliseconds=RecordDetails->Value.TimeStamp.tv_usec/1000;
            Fieldp->ValueTime = Fieldp->ValueTime      /* update timer totals */
                             + (ElapsedTime*1000) + milliseconds;
 
            if (SummaryRF == TRUE) {                  /* set this row's timer */
               for(i=0; i<FieldCount; i++) {
                  if (RecordDetails->FieldID == row->ItemS[i].Fieldp->empNum) {
                     row->ItemS[i].RowTime = (ElapsedTime*1000) + milliseconds;
                     break;
                  }
               }
            }

            if (TimeInSec) {
               sprintf(wrkbuf, "%8d.%.03d", ElapsedTime, milliseconds);
            } else {
               seconds=ElapsedTime%60;
               total_minutes=ElapsedTime/60;
               minutes=total_minutes%60;
               total_hours=total_minutes/60;
               hours=total_hours%24;
               memset(HH, '\0', 1);                      /* set hours to null */
               memset(MM, '\0', 1);                    /* set minutes to null */
               memset(SS, '\0', 1);                    /* set seconds to null */
               if (hours != 0)
                       sprintf(HH, "%02d:", hours);          /* display hours */
               if (hours != 0 || minutes != 0)
                       sprintf(MM, "%02d:", minutes);      /* display minutes */
               if (hours != 0 || minutes != 0 || seconds != 0)
                       sprintf(SS, "%02d", seconds);       /* display seconds */
               sprintf(MS, "%03d", milliseconds);     /* display milliseconds */
               sprintf(wrkbuf, "%s%s%s.%s", HH, MM, SS, MS);      /* build it */
               if (! memcmp(&wrkbuf[0], "0", 1))
                   memset(&wrkbuf[0], ' ', 1);          /* leading 0 to blank */
            }
            sprintf(fieldValue, "%*.*s", fldSize, fldSize, wrkbuf); /* return */
            break;

        case CICS_EMP_TDQ_DETAIL_TYPE_TIMER:
            /*---------------------------------------------------------*/
            /* Basic Time Stamp with Counter ...                       */
            /* Elapsed time in value one word format: CICS_EMP_Clock_t */
            /* convert this to hh:mm:ss.ms  & clock count              */ 
            /* The timer accuracy for this field is nanoseconds        */
            /*---------------------------------------------------------*/
#ifdef DEBUG2
            fprintf(stderr, "[MonExpand]: TIMER(%i) %s\n",
                    fldSize, Fieldp->fieldName);
#endif
            ElapsedTime=RecordDetails->Value.Timer.TotalTime/1000000;
            milliseconds=RecordDetails->Value.Timer.TotalTime%1000000;   
            milliseconds=milliseconds/1000;   
            Fieldp->ValueTime = Fieldp->ValueTime      /* update timer totals */
                             + (ElapsedTime*1000) + milliseconds;
            Fieldp->ValueCtr = Fieldp->ValueCtr      /* update counter totals */
                             + RecordDetails->Value.Timer.Counter;

            if (SummaryRF == TRUE) {          /* save this row's accumulators */
               for(i=0; i<FieldCount; i++) {
                  if (RecordDetails->FieldID == row->ItemS[i].Fieldp->empNum) {
                     row->ItemS[i].RowTime = (ElapsedTime*1000) + milliseconds;
                     row->ItemS[i].RowCount =
                                           RecordDetails->Value.Timer.Counter;
                     break;
                  }
               }
            }

            if (TimeInSec)
            {
               sprintf(wrkbuf, "%8d.%.03d", ElapsedTime, milliseconds);
            } else {
               seconds=ElapsedTime%60;
               total_minutes=ElapsedTime/60;
               minutes=total_minutes%60;
               total_hours=total_minutes/60;
               hours=total_hours%24;
               memset(HH, '\0', 1);                      /* set hours to null */
               memset(MM, '\0', 1);                    /* set minutes to null */
               memset(SS, '\0', 1);                    /* set seconds to null */
               if (hours != 0)
                       sprintf(HH, "%02d:", hours);          /* display hours */
               if (hours != 0 || minutes != 0)
                       sprintf(MM, "%02d:", minutes);      /* display minutes */
               if (hours != 0 || minutes != 0 || seconds != 0)
                       sprintf(SS, "%02d", seconds);       /* display seconds */
               sprintf(MS, "%03d", milliseconds);     /* display milliseconds */
               sprintf(wrkbuf, "%s%s%s.%s", HH, MM, SS, MS);      /* build it */
               if (! memcmp(&wrkbuf[0], "0", 1))
                   memset(&wrkbuf[0], ' ', 1);          /* leading 0 to blank */
            }
            sprintf(fieldValue, "%*.*s", fldSize, fldSize, wrkbuf); /* return */
            if (PrtCount==TRUE) {
               sprintf(wrkbuf, "(%05d)", RecordDetails->Value.Timer.Counter);
               strcat(fieldValue, wrkbuf);               /* add counter value */
            }
            break;

        case CICS_EMP_TDQ_DETAIL_TYPE_HIGH_WATER:
            /*--------------------------------------------*/
            /* Storage High Watermark Values ...          */
            /* CICS_EMP_HighWater = u_long_t              */
            /* convert bytes HWM to kbytes HWM   u_long_t */
            /*--------------------------------------------*/
#ifdef DEBUG2
            fprintf(stderr, "[MonExpand]: HIGH_WATER(%i) %s\n",
                    fldSize, Fieldp->fieldName);
#endif
            sprintf(wrkbuf, "%.1dk", RecordDetails->Value.Counter/1024);
            sprintf(fieldValue, "%*.*s", fldSize, fldSize, wrkbuf); /* return */
            break;

        case CICS_EMP_TDQ_DETAIL_TYPE_OCCUPANCY:
            /*------------------------------------------------------*/
            /* Storage Occupancy (usec times kbytes)                */
            /* CICS_EMP_Occupancy = u_xlong_t                       */
            /* this is double-long bytes per micsec: 2@32 bits each */ 
            /* convert it to seconds times kilobytes                */ 
            /*------------------------------------------------------*/
#ifdef DEBUG2
            fprintf(stderr, "[MonExpand]: OCCUPANCY(%i) %s\n",
                    fldSize, Fieldp->fieldName);
#endif
            fpointnum=RecordDetails->Value.Occupancy.High * 4294967296.0 +
                      RecordDetails->Value.Occupancy.Low ;  
            fpointnum=fpointnum/1000000;             /* bytes used per second */
            fpointnum=fpointnum/1024;               /* kbytes used per second */
            sprintf(wrkbuf, "%.3f", fpointnum);
            sprintf(fieldValue, "%*.*s", fldSize, fldSize, wrkbuf); /* return */
            break;

       case CICS_EMP_TDQ_DETAIL_TYPE_ABEND_CODE:
            /*-----------------------------------*/
            /* Abend Codes ...                   */
            /*-----------------------------------*/
#ifdef DEBUG2
            fprintf(stderr, "[MonExpand]: ABEND_CODE(%i) %s\n",
                    fldSize, Fieldp->fieldName);
#endif
            memcpy(wrkbuf, RecordDetails->Value.AbendCode, CICS_EMP_ABEND_SIZE);
            wrkbuf[CICS_EMP_ABEND_SIZE]='\0';
            sprintf(fieldValue, "%*.*s", fldSize, fldSize, wrkbuf); /* return */

            if (SummaryRF == TRUE) {                /* save displayable value */
               if (Table.Major->empNum == RecordDetails->FieldID)
                  strcpy(row->MajorValue, fieldValue);
               if (Table.Minor != NULL)
                  if (Table.Minor->empNum == RecordDetails->FieldID)
                     strcpy(row->MinorValue, fieldValue);
            }

            break;

        case CICS_EMP_TDQ_DETAIL_TYPE_USER:
            /*-------------------------------------*/
            /* User Defined Data Type              */
            /*-------------------------------------*/
#ifdef DEBUG2
            fprintf(stderr, "[MonExpand]: USER(%i) %s\n",
                    fldSize, Fieldp->fieldName);
#endif
            empPtr = (struct CICS_EMP_Record *)&RecordDetails->Value;
            switch(Fieldp->usrType)                /* based on it's user type */
            {
            case Clk1T:
                 ElapsedTime=empPtr->Clk_Time1/1000000;
                 milliseconds=empPtr->Clk_Time1%1000000;   
                 milliseconds=milliseconds/1000;   
                 Fieldp->ValueTime = Fieldp->ValueTime
                     + (ElapsedTime*1000) + milliseconds;
                 if (TimeInSec) {
                    sprintf(wrkbuf, "%8d.%.03d", ElapsedTime, milliseconds);
                 } else {
                    seconds=ElapsedTime%60;
                    total_minutes=ElapsedTime/60;
                    minutes=total_minutes%60;
                    total_hours=total_minutes/60;
                    hours=total_hours%24;
                    memset(HH, '\0', 1);                 /* set hours to null */
                    memset(MM, '\0', 1);               /* set minutes to null */
                    memset(SS, '\0', 1);               /* set seconds to null */
                    if (hours != 0)
                            sprintf(HH, "%02d:", hours);     /* display hours */
                    if (hours != 0 || minutes != 0)
                            sprintf(MM, "%02d:", minutes); /* display minutes */
                    if (hours != 0 || minutes != 0 || seconds != 0)
                            sprintf(SS, "%02d", seconds);  /* display seconds */
                    sprintf(MS, "%03d", milliseconds);/* display milliseconds */
                    sprintf(wrkbuf, "%s%s%s.%s", HH, MM, SS, MS); /* build it */
                    if (! memcmp(&wrkbuf[0], "0", 1))
                        memset(&wrkbuf[0], ' ', 1);     /* leading 0 to blank */
                 }
                 break;
            case Clk1Ct:
                 Fieldp->ValueCtr = Fieldp->ValueCtr + empPtr->Clk_Counter1;
                 sprintf(wrkbuf, "%ld", empPtr->Clk_Counter1);
                 break;
            case Clk2T:
                 ElapsedTime=empPtr->Clk_Time2/1000000;
                 milliseconds=empPtr->Clk_Time2%1000000;   
                 milliseconds=milliseconds/1000;   
                 Fieldp->ValueTime = Fieldp->ValueTime
                     + (ElapsedTime*1000) + milliseconds;
                 if (TimeInSec) {
                    sprintf(wrkbuf, "%8d.%.03d", ElapsedTime, milliseconds);
                 } else {
                    seconds=ElapsedTime%60;
                    total_minutes=ElapsedTime/60;
                    minutes=total_minutes%60;
                    total_hours=total_minutes/60;
                    hours=total_hours%24;
                    memset(HH, '\0', 1);                 /* set hours to null */
                    memset(MM, '\0', 1);               /* set minutes to null */
                    memset(SS, '\0', 1);               /* set seconds to null */
                    if (hours != 0)
                            sprintf(HH, "%02d:", hours);     /* display hours */
                    if (hours != 0 || minutes != 0)
                            sprintf(MM, "%02d:", minutes); /* display minutes */
                    if (hours != 0 || minutes != 0 || seconds != 0)
                            sprintf(SS, "%02d", seconds);  /* display seconds */
                    sprintf(MS, "%03d", milliseconds);/* display milliseconds */
                    sprintf(wrkbuf, "%s%s%s.%s", HH, MM, SS, MS); /* build it */
                    if (! memcmp(&wrkbuf[0], "0", 1))
                        memset(&wrkbuf[0], ' ', 1);     /* leading 0 to blank */
                 }
                 break;
            case Clk2Ct:
                 Fieldp->ValueCtr = Fieldp->ValueCtr + empPtr->Clk_Counter2;
                 sprintf(wrkbuf, "%ld", empPtr->Clk_Counter2);
                 break;
            case Ctr1:
                 Fieldp->ValueCtr = Fieldp->ValueCtr + empPtr->Counter1r;
                 sprintf(wrkbuf, "%ld", empPtr->Counter1r);
                 break;
            case Ctr2:
                 Fieldp->ValueCtr = Fieldp->ValueCtr + empPtr->Counter2r;
                 sprintf(wrkbuf, "%ld", empPtr->Counter2r);
                 break;
            case Ctr3:
                 Fieldp->ValueCtr = Fieldp->ValueCtr + empPtr->Counter3r;
                 sprintf(wrkbuf, "%ld", empPtr->Counter3r);
                 break;
            default:
                 sprintf(fieldValue, "<-User->");
                 break;
            }
            sprintf(fieldValue, "%*.*s", fldSize, fldSize, wrkbuf);
            break;

        default:
            /*-----------------------------------*/
            /* Error, unknown or unsupported ... */
            /*-----------------------------------*/
            sprintf(fieldValue, "<?EMP??>");
            break;
    }
    return(0);
} /* end MonExpand() */

/******************************************************************************/
/* MonTitle()                     Create title line based on selected fields  */
/******************************************************************************/
int   MonTitle(CICS_EMP_TDQ_RecordDetails_t *RecordDetails,
               struct FieldDesc_t           *Fieldp,
               char                          titleTxt[])
{
  int    i, rc;                                               /* work integer */
  char   FldDesc[128]={'\0'};                 /* EMP Report Field Description */

  i=(Fieldp->outSize);                          /* get the report field width */
  rc=0;                                                          /* OK return */

  /*----------------------------------------*/
  /* Process CMF User 2 Clocks & 3 Counters */
  /*----------------------------------------*/
  if (RecordDetails->FieldID == CICS_EMP_FID_USER_SPACE
      && RecordDetails->Type == CICS_EMP_TDQ_DETAIL_TYPE_USER
      && Fieldp->usrType == none)
  {
    sprintf(titleTxt, "%*.*s", i, i,                             /* format it */
      " UserClock-1  count  UserClock-2  count UsrCt1 UsrCt2 UsrCt3");
    return(rc);
  }

  /*----------------------------------------*/
  /* Process all other fields ...           */
  /*----------------------------------------*/
  if (RecordDetails == NULL)                      /* Invalid Field Selected ? */
  {
    sprintf(FldDesc, "%s", Fieldp->titleTxt);                    /* format it */
    rc=1;                                                  /* non zero return */
  } else {
    sprintf(FldDesc, "%*.*s", i, i, Fieldp->titleTxt);           /* format it */
    if (PrtCount && RecordDetails->Type == CICS_EMP_TDQ_DETAIL_TYPE_TIMER)
        strcat(FldDesc, " count ");           /* set area for the clock count */
  }
  sprintf(titleTxt, "%s", FldDesc);          /* return title string to caller */

  return(rc);
} /* end MonTitle() */

/******************************************************************************/
/* MonCompare()                         Select records based on field values  */
/******************************************************************************/
int MonCompare (CICS_EMP_TDQ_RecordDetails_t *RecordDetails,
                struct ExcIncList_t           CompareItem)
{
  int       rc, rcComp;                     /* return value for compared data */
  long      TimeValue;                          /* time calculation work area */
  struct tm TimeData;                                 /* time value work area */
  time_t    TimeTOD;                                  /* time value work area */

    if (RecordDetails == NULL)                    /* Invalid Field Selected ? */
        return(FALSE);

    switch(RecordDetails->Type)            /* test a field based on it's type */
    {
        /*-----------------------*/
        /* Character Strings ... */
        /*-----------------------*/
        case CICS_EMP_TDQ_DETAIL_TYPE_STRING:
#ifdef DEBUG3
            printf(stderr, "comparing (%d)\"%s\" ",
                  strlen(CompareItem.text),
                  CompareItem.text);
            switch (CompareItem.oper)
            {
               case GREATERTHAN:
                    fprintf(stderr, ">"); break;
               case LESSTHAN:
                    fprintf(stderr, "<"); break;
               case EQUALS:
                    fprintf(stderr, "="); break;
               default:
                    fprintf(stderr, "?"); break;
            }
            fprintf(stderr, " (%d)\"%0.8s\"\n",
                   RecordDetails->FieldSize,
                   RecordDetails->Value.String);
#endif
            rcComp = memcmp(RecordDetails->Value.String,
                            CompareItem.text,
                            strlen(CompareItem.text));     /* compare strings */
#ifdef DEBUG3
            fprintf(stderr, "memcmp rcComp(%i)\n", rc);
#endif
            rc=FALSE;                                          /* if no match */
            switch (CompareItem.oper) {
               case GREATERTHAN:
                    if (rcComp > 0)  rc=TRUE;           /* greater than match */
                    break;
               case LESSTHAN:
                    if (rcComp < 0)  rc=TRUE;              /* less than match */
                    break;
               case EQUALS:
                    if (rcComp == 0) rc=TRUE;                  /* equal match */
                    break;
               default:
                    return(FALSE);                            /* no match ... */
                    break;
            }
#ifdef DEBUG3
            fprintf(stderr, "Returning rc(%i)\n", rc);
#endif
            return(rc);
            break;

        /*-----------------------*/
        /* Basic Counters    ... */
        /*-----------------------*/
        case CICS_EMP_TDQ_DETAIL_TYPE_OS_COUNT:
        case CICS_EMP_TDQ_DETAIL_TYPE_COUNTER:
            if     (RecordDetails->Value.Counter < CompareItem.value) rcComp=-1;
            else if(RecordDetails->Value.Counter > CompareItem.value) rcComp=+1;
            else                                                      rcComp=0;

#ifdef DEBUG3
            printf(stderr, "if Counter(%i) = value(%i) rc(%i)\n",
                   RecordDetails->Value.Counter, CompareItem.value, rcComp);
#endif
            rc=FALSE;                                          /* if no match */
            switch (CompareItem.oper) {
               case GREATERTHAN:
                    if (rcComp > 0)  rc=TRUE;           /* greater than match */
                    break;
               case LESSTHAN:
                    if (rcComp < 0)  rc=TRUE;              /* less than match */
                    break;
               case EQUALS:
                    if (rcComp == 0) rc=TRUE;                  /* equal match */
                    break;
               default:
                    return(FALSE);                            /* no match ... */
                    break;
            }
#ifdef DEBUG3
            fprintf(stderr, "Returning rc(%i)\n", rc);
#endif
            return(rc);
            break;

        /*-----------------------*/
        /* Unix Clocks       ... */
        /*-----------------------*/
        case CICS_EMP_TDQ_DETAIL_TYPE_OS_CLOCK:
            if     (RecordDetails->Value.TimeStamp.tv_sec
                    < CompareItem.value)                         rcComp=-1;
            else if(RecordDetails->Value.TimeStamp.tv_sec
                    > CompareItem.value)                         rcComp=+1;
            else                                                 rcComp=0;

#ifdef DEBUG3
            fprintf(stderr, "if Time_seconds (%i) = value(%i) rc(%i)\n",
                    RecordDetails->Value.TimeStamp.tv_sec,
                    CompareItem.value, rcComp);
#endif
            rc=FALSE;                                          /* if no match */
            switch (CompareItem.oper) {
               case GREATERTHAN:
                    if (rcComp > 0)  rc=TRUE;           /* greater than match */
                    break;
               case LESSTHAN:
                    if (rcComp < 0)  rc=TRUE;              /* less than match */
                    break;
               case EQUALS:
                    if (rcComp == 0) rc=TRUE;                  /* equal match */
                    break;
               default:
                    return(FALSE);                            /* no match ... */
                    break;
            }
#ifdef DEBUG3
            fprintf(stderr, "Returning rc(%i)\n", rc);
#endif
            return(rc);
            break;

        /*-----------------------*/
        /* CICS Clocks       ... */
        /*-----------------------*/
        case CICS_EMP_TDQ_DETAIL_TYPE_TIMER:
            TimeValue = RecordDetails->Value.Timer.TotalTime/1000000;
            if     (TimeValue < CompareItem.value)             rcComp=-1;
            else if(TimeValue > CompareItem.value)             rcComp=+1;
            else                                               rcComp=0;

#ifdef DEBUG3
            fprintf(stderr, "if Time_seconds (%i) = value(%i) rc(%i)\n",
                    TimeValue, CompareItem.value, rcComp);
#endif
            rc=FALSE;                                          /* if no match */
            switch (CompareItem.oper) {
               case GREATERTHAN:
                    if (rcComp > 0)  rc=TRUE;           /* greater than match */
                    break;
               case LESSTHAN:
                    if (rcComp < 0)  rc=TRUE;              /* less than match */
                    break;
               case EQUALS:
                    if (rcComp == 0) rc=TRUE;                  /* equal match */
                    break;
               default:
                    return(FALSE);                            /* no match ... */
                    break;
            }
#ifdef DEBUG3
            fprintf(stderr, "Returning rc(%i)\n", rc);
#endif
            return(rc);
            break;

        /*-----------------------*/
        /* TOD Timestamps    ... */
        /*-----------------------*/
        case CICS_EMP_TDQ_DETAIL_TYPE_TIME_STAMP:
            TimeTOD = RecordDetails->Value.TimeStamp.tv_sec;
            TimeTOD += RecordDetails->Value.TimeStamp.tv_usec/1000000;
            if (GMT) {
               memcpy(&TimeData, gmtime(&TimeTOD), sizeof(struct tm));
            } else {
               memcpy(&TimeData, localtime(&TimeTOD), sizeof(struct tm));
            }
            TimeValue = (TimeData.tm_hour*3600)
                        + (TimeData.tm_min*60)
                        +  TimeData.tm_sec;

            if     (TimeValue < CompareItem.value)             rcComp=-1;
            else if(TimeValue > CompareItem.value)             rcComp=+1;
            else                                               rcComp=0;

#ifdef DEBUG3
            printf(stderr, "if TOD (%i) = value(%i) rc(%i)\n",
                   TimeValue, CompareItem.value, rcComp);
#endif
            rc=FALSE;                                          /* if no match */
            switch (CompareItem.oper) {
               case GREATERTHAN:
                    if (rcComp > 0)  rc=TRUE;           /* greater than match */
                    break;
               case LESSTHAN:
                    if (rcComp < 0)  rc=TRUE;              /* less than match */
                    break;
               case EQUALS:
                    if (rcComp == 0) rc=TRUE;                  /* equal match */
                    break;
               default:
                    return(FALSE);                            /* no match ... */
                    break;
            }
#ifdef DEBUG3
            fprintf(stderr, "Returning rc(%i)\n", rc);
#endif
            return(rc);
            break;

        /*-------------------------------------*/
        /* Error, unknown or unsupported field */
        /*-------------------------------------*/
        default:
            return(FALSE);
            break;
    }
    return(FALSE);                                            /* no match ... */
} /* end MonCompare() */

/******************************************************************************/
/* insertRow()      Insert a SummaryRow into the global table.                */
/*   Accumulate totals for the row if a match (major/minor) is found,         */
/*   else, create a new row for this major/minor pair.                        */
/*   A non-zero return value indicates an error.                              */
/******************************************************************************/
int insertRow(struct SummaryRow_t* row)
{
  int    i, MajorC, MinorC;
  struct SummaryRow_t* ptr;
  struct SummaryRow_t* backptr;
  struct SummaryRow_t* newptr;

   /*-------------------------------------------------------------*/
   /* Initialize local variables.  ptr will point to a row in the */
   /* table while backptr will point to the previous row or NULL  */
   /* when ptr points to the first row.                           */
   /*-------------------------------------------------------------*/
   ptr = Table.Head;
   backptr = NULL;

   /*---------------------------------------------*/
   /* Iterate through the Table,                  */
   /* comparing each table row with the input row */
   /*---------------------------------------------*/
   while (ptr != NULL) {
      MajorC = strcmp(ptr->MajorValue, row->MajorValue);
      MinorC = strcmp(ptr->MinorValue, row->MinorValue);
      /*------------------------------------------------------------------*/
      /* When the Major and Minor columns have an exact match, accumulate */
      /*------------------------------------------------------------------*/
      if ((MajorC == 0) && (MinorC == 0)) {
         ptr->ItemCount++;
         for(i=0; i<FieldCount; i++) {
            ptr->ItemS[i].RowCount += row->ItemS[i].RowCount;
            ptr->ItemS[i].RowTime += row->ItemS[i].RowTime;
         }
         return(0);
      }

      /*----------------------------------------------------------------*/
      /* When the Major columns match but the Minor columns do not then */
      /* create a new row when the table row's Minor column is larger   */
      /* than this row's Minor column.  If the Major columns do not     */
      /* match and the table row's Major column is larger then this     */
      /* row's Major column then create a new row.  In any case, break  */
      /* out of the loop to retain the iterating pointer values.        */
      /*----------------------------------------------------------------*/
      if (((MajorC == 0) && (MinorC > 0)) || (MajorC > 0))
         break;

      backptr = ptr;                               /* Advance to the next row */
      ptr = ptr->Next;
   }

   /*-----------------------------------------------------------*/
   /* An exact match was not found.  A new row must be created. */
   /*-----------------------------------------------------------*/
   newptr = (struct SummaryRow_t*) malloc(sizeof(struct SummaryRow_t));
   if (newptr == NULL) {
      fprintf(stderr, "Memory allocation error!\n");
      return(1);
   }

   newptr->ItemS = (struct SummaryItem_t*) calloc(FieldCount,
                    sizeof(struct SummaryItem_t));
   if (newptr->ItemS == NULL) {
      fprintf(stderr, "Memory allocation error!\n");
      return(1);
   }

   strcpy(newptr->MajorValue, row->MajorValue);
   strcpy(newptr->MinorValue, row->MinorValue);
   newptr->ItemCount = 1;
   for(i=0; i<FieldCount; i++) {
      newptr->ItemS[i].Fieldp = row->ItemS[i].Fieldp;
      newptr->ItemS[i].RowCount = row->ItemS[i].RowCount;
      newptr->ItemS[i].RowTime = row->ItemS[i].RowTime;
   }

   /*---------------------*/
   /* Adjust the pointers */
   /*---------------------*/
   if (backptr == NULL) {                         /* case 1: insert first row */
      newptr->Next = ptr;
      Table.Head = newptr;
   } else {                              /* case 2: insert after back pointer */
      newptr->Next = ptr;
      backptr->Next = newptr;
   }

   return(0);
}
